from .action_status import ActionStatus
from .action_tracker import ActionTracker
from .base_path_structure import BasePathStructure
from .manager import WorkspaceManager
from .state import WorkspaceState
from .storage import WorkspaceStorage
from .workspace import Workspace
from .path_structure_meta import PathStructureMeta
from .bind import Bind
from .base_data import BaseData


__version__ = '4.1.2'

__all__ = [
    'Workspace',
    'WorkspaceState',
    'WorkspaceStorage',
    'WorkspaceManager',
    'ActionTracker',
    'ActionStatus',
    'BasePathStructure',
    'BaseData',
    'Bind',
]



import logging
import os

from .state import WorkspaceState
from .bind import Bind
from .action_status import ActionStatus


def _get_logger():
    return logging.getLogger(__name__)


class ActionTracker:
    """Manages and monitors actions"""

    def __init__(self, action: str, state_path: str, path=None):
        """Creates a new action tracker

        Attributes
        ----------
        action: str
            Name of the action to track
        state_path: str
            Path of the state file to store information about the action state
        path: str, optional
            Path of the file to check for existatnce on exit
            
            This path is checked only if the tracker 
            is used as a context manager
        """
        assert isinstance(action, str)
        assert isinstance(state_path, str)

        self.action = action
        self.state_path = state_path
        self.path = path

    def __repr__(self):
        return '<ActionTracker({})>'.format(self.action)

    @property
    def state(self) -> WorkspaceState:
        """WorkspaceState object linked to `self.state_path`"""
        return WorkspaceState(self.state_path)

    def __enter__(self):
        """Context manager enter"""
        self.state[self.action] = ActionStatus.started.value

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit"""
        there_is_no_exception = exc_tb is None
        file_exists = self.path is None or os.path.exists(self.path)

        status = (
            ActionStatus.finished
            if there_is_no_exception and file_exists
            else ActionStatus.failed
        )

        if status is ActionStatus.failed:
            if not there_is_no_exception:
                _get_logger().info('Fail cause: {}'.format(exc_tb))
            elif not file_exists:
                _get_logger().info("File doesn't exist: {}".format(self.path))

        self.state[self.action] = status.value

    def _status_flag(self, status):
        return self.status() == status.value

    def started(self):
        """Returns True if the action has started and is pending"""
        return self._status_flag(ActionStatus.started)

    def finished(self):
        """Returns True if the action has finished"""
        return self._status_flag(ActionStatus.finished)

    def failed(self):
        """Returns True if the action has failed"""
        return self._status_flag(ActionStatus.failed)

    def undefined(self):
        """Returns True if the action state is not defined"""
        return self.state[self.action] is None

    def status(self):
        """Returns a string representation of the current action status"""
        return self.state[self.action]



import logging

import os

from .loaders import default_loader
from .bind import Bind


def _get_logger():
    return logging.getLogger(__name__)


def _get_path_property(name):

    def _joined_path(self):
        bind_obj = self._path_meta[name].path  # type: Bind
        abs_path = os.path.join(self._root, bind_obj)

        action = bind_obj.action()

        tracker = self.track(action)

        msg = "Action '{}' status: {}".format(
            action,
            tracker.status(),
        )
        _get_logger().info(msg)

        if tracker.finished():
            return Bind(abs_path)

        if not tracker.started():
            _get_logger().info("Starting action: '{}'".format(action))
            bind_obj.loader(
                str(abs_path),
                str(self._root),
                str(action),
            )

            msg = "Action '{}' status: {}".format(action, tracker.status())
            _get_logger().info(msg)

            if tracker.finished():
                return Bind(abs_path)
        else:
            _get_logger().info("Path not ready: '{}' in progress".format(action))

        return None

    return property(_joined_path)


class CustomProperty(property):
    def __init__(self, p, action):
        super().__init__(p.fget, p.fset, p.fdel)
        self.action = action


def _get_loader(value):

    if isinstance(value, Bind):
        loader = value.loader
    elif isinstance(value, str):
        loader = default_loader
    else:
        raise ValueError('PathStructure value type must be one of {}, found: {}'.format(
            (Bind, str), type(value)
        ))

    return loader


def _handle_attribute(name, value):
    if name.startswith('_') or name == 'track':
        return {}

    loader = _get_loader(value)

    meta = _PathItemMeta(
        Bind(str(value), loader),
        _get_path_property(name),
    )
    meta.prop = CustomProperty(meta.prop, meta.path.action)
    return meta


class _PathItemMeta:
    def __init__(self, path, prop):
        self.path = path
        self.prop = prop


class PathStructureMeta(type):
    def __new__(mcs, cls_name, bases, attrib):
        paths = {}

        for base in bases:
            _paths = getattr(base, '_path_meta')
            paths.update(_paths)

        for name, value in attrib.items():
            meta = _handle_attribute(name, value)
            if meta:
                paths.update({name: meta})

        attrib['_path_meta'] = paths

        for path, meta in paths.items():
            attrib[path] = meta.prop

        return super().__new__(mcs, cls_name, bases, attrib)






class _DataInfo:
    def __init__(self, loader):
        self.cached = None
        self.loader = loader


class _DataItemMeta:
    def __init__(self, info, prop):
        self.info = info
        self.prop = prop


def _get_data_property(name):

    def _cached_data(self):
        info = self._data_meta[name].info  # type: _DataInfo
        if info.cached is None:
            info.cached = info.loader(self)
        return info.cached

    return property(_cached_data)


def _cycle(name, value):
    if name.startswith('_'):
        return None

    return _DataItemMeta(
        _DataInfo(value),
        _get_data_property(name),
    )


class DataMeta(type):
    """Meta class for BaseData subclasses"""
    def __new__(mcs, cls_name, bases, attrib):
        items = {}

        for base in bases:
            items.update(getattr(base, '_data_meta'))

        for name, value in attrib.items():
            item = _cycle(name, value)
            if item:
                items[name] = item

        attrib['_data_meta'] = items

        for name, item in items.items():
            attrib[name] = item.prop

        return super().__new__(mcs, cls_name, bases, attrib)




class Bind(str):
    """String wrapper class to hold a loader attribute
    
    Attributes
    ----------
    path: str
    loader: function
    """

    def __new__(cls, path, *args, **kwargs):
        return super().__new__(cls, path)

    def __init__(self,
                 path: str,
                 loader=None):
        """Creates a path with a loader attribute

        Attributes
        ----------
        path: str
            Target path
        loader: function
            Loader function will be called 
            the first time the path is accessed
        """
        self.path = path
        self.loader = loader

    def action(self):
        """Construct an action name from a Bind object"""
        return str(self) + '_load'

    @classmethod
    def from_action(cls, action):
        """Construct a Bind object from an action name"""
        return (
            Bind(action[:-5], None)
            if action[-5:] == '_load'
            else None
        )



import os


def create_dir(path):
    os.makedirs(path, exist_ok=True)


def create_file(path):
    if not os.path.exists(path):
        create_dir(os.path.split(path)[0])
        with open(path, 'w'):
            pass

def is_dir(path):
    name = os.path.split(path)[-1]  # type: str
    stripped_name = name.lstrip('.')
    return '.' not in stripped_name


def default_loader(path:str, root:str, action:str):
    from .base_path_structure import BasePathStructure

    ws = BasePathStructure(root)
    with ws.track(action):
        if is_dir(path):
            create_dir(path)
        else:
            create_file(path)



from .base_path_structure import BasePathStructure
from .data_meta import DataMeta


class BaseData(metaclass=DataMeta):
    """Base class for the data

    Examples
    --------

    :ref:`basic-data-loader`

    :ref:`pass-data-loader-args`

    :ref:`chained-data-loaders`
    """

    def __init__(self, path_structure=None):
        """Creates a Data instance

        Attributes
        ----------
        path_structure: BasePathStructure
        """
        self._path = path_structure
        if self._path:
            self._path._bind_data(self)

    def __repr__(self):
        return '<{} path={}>'.format(
            self.__class__.__name__,
            self._path
        )


import os

from .base_path_structure import BasePathStructure
from .action_tracker import ActionTracker
from .base_data import BaseData


class Workspace:
    """Wrapper interface to store path and data
    
    Attributes
    ----------
    id: int or str
    root: str
    path: BasePathStructure
    data: BaseData
    
    Examples
    --------

    :ref:`basic-workspace`

    :ref:`workspace-with-chained-loaders`

    """

    def __init__(self, data):
        """Creates a Workspace object"""
        assert isinstance(data, BaseData)
        self.data = data

    @property
    def path(self) -> BasePathStructure:
        """Bound path structure object"""
        return self.data._path

    @property
    def root(self) -> str:
        """Workspace root"""
        return self.path._root

    def __repr__(self):
        return '<{}({})>'.format(self.__class__.__name__, self.root)

    @property
    def id(self):
        """Workspace id

        Id is used to sort workspaces from oldest to latest
        """
        dirname = os.path.split(self.root)[-1]
        try:
            return int(dirname)
        except ValueError:
            return str(dirname)

    def track(self, action: str) -> ActionTracker:
        """Creates an ActionTracker object to manage action `action`

        It's a shortcut to avoid creating ActionTracker directly

        Attributes
        ----------
        action: str
            Action to track
        """
        return self.path.track(action)

    def __eq__(self, other):
        return self.root == other.root

    @classmethod
    def construct(cls, root, path_structure_cls=None, data_cls=None):
        """Creates a Workspace object
        
        Convinient alternative constructor
        """
        if not path_structure_cls:
            path_structure_cls = BasePathStructure
        if not data_cls:
            data_cls = BaseData
        return Workspace(data_cls(path_structure_cls(root)))




import json
import os


def _store_state(state, path):
    with open(path, 'w') as f:
        json.dump(state, f, indent=2)
        f.write('\n')


def _load_state(path):
    try:
        JSONDecodeError = json.decoder.JSONDecodeError
    except AttributeError:
        JSONDecodeError = ValueError
        
    try:
        with open(path, 'r') as f:
            state = json.load(f)
    except JSONDecodeError:
        state = {}

    return state


class WorkspaceState:
    """Manages workspace state

    State is written to file each time anything changes
    
    Attributes
    ----------
    path: str
        Path of the file to store the state in
    """

    def __init__(self, path: str):
        """Creates a WorkspaceState object
        
        Attributes
        ----------
        path: str
            Path of the file to store the state in
        """
        self.path = str(path)

        if not os.path.exists(self.path):
            state = {}
            _store_state(state, self.path)

    def __getitem__(self, item):
        state = _load_state(self.path)
        return (
            state[item]
            if item in state
            else None
        )

    def __setitem__(self, key, value):
        path = self.path
        state = _load_state(path)
        state[key] = value
        _store_state(state, path)

    def __iter__(self):
        return iter(_load_state(self.path))



import os
import warnings
from typing import Optional, Type

from .base_path_structure import BasePathStructure
from .state import WorkspaceState
from .storage import WorkspaceStorage
from .workspace import Workspace
from .base_data import BaseData
from .constants import _STATE, _CURRENT, _TARGET


class WorkspaceManager:
    """Manages workspaces

    Allows you to
    - create workspaces
    - load workspaces from directory
    - manage current, target and latest workspaces
    - search for workspaces with finished actions

    Examples
    --------

    :ref:`workspace-management`

    :ref:`manager-with-data-bindings`
    
    """

    def __init__(self,
                 root,
                 path_structure_cls=None,
                 data_cls=None,
                 workspace_cls=None):
        """Creates a workspace manager

        Parameters
        ----------
        root: str
            Workspace storage path
        path_structure_cls: BasePathStructure class object
        data_cls: BaseData class object
        workspace_cls: Workspace class object
        """
        self.root = str(root)
        os.makedirs(self.root, exist_ok=True)

        self.workspace_cls = (
            workspace_cls
            if workspace_cls
            else Workspace
        )
        self.path_structure_cls = (
            path_structure_cls
            if path_structure_cls
            else BasePathStructure
        )
        self.data_cls = (
            data_cls
            if data_cls
            else BaseData
        )

        self.storage = WorkspaceStorage.load_from_directory(
            self.root,
            path_structure_cls=self.path_structure_cls,
            data_cls=self.data_cls,
            workspace_cls=self.workspace_cls,
        )

        self.state = WorkspaceState(os.path.join(self.root, _STATE))

    @property
    def current_ws_id(self):
        """Returns current workspace id"""
        return self.state[_CURRENT]

    @current_ws_id.setter
    def current_ws_id(self, value):
        """Sets current workspace id"""
        self.state[_CURRENT] = value

    @property
    def target_ws_id(self):
        """Returns target workspace id"""
        return self.state[_TARGET]

    @target_ws_id.setter
    def target_ws_id(self, value):
        """Sets target workspace id"""
        self.state[_TARGET] = value

    def create(self, ws_id) -> Workspace:
        """Creates a Workspace

        Parameters
        ----------
        ws_id: str or int
            Workspace id
        """
        name = str(ws_id)

        ws_root = os.path.join(self.root, name)
        ws = self.workspace_cls.construct(
            ws_root,
            self.path_structure_cls,
            self.data_cls,
        )
        self.storage.add(ws)
        return ws

    def latest(self) -> Workspace:
        """Returns latest workspace"""
        return self.storage.at(-1)

    def current(self) -> Optional[Workspace]:
        """Returns current workspace"""
        return self.storage[self.current_ws_id]

    def target(self) -> Workspace:
        """Returns target workspace"""
        return self.storage[self.target_ws_id]

    def __repr__(self):
        return '<{}(root={}, storage={})>'.format(
            self.__class__.__name__,
            self.root,
            self.storage,
        )

    def find_latest_finished(self, action) -> Optional[Workspace]:
        """Searches for latest workspace with finished action `action`"""
        workspaces = list(self.storage.values())
        for ws in reversed(workspaces):
            tracker = ws.track(action)
            if tracker.finished():
                return ws

    def target_latest(self):
        """Sets the target id equal to the latest id"""
        self.target_ws_id = self.latest().id

    def update(self):
        """Migrates from current workspace to target one
        
        Sets the current workspace id equal to the target workspace id
        """
        if self.current_ws_id == self.target_ws_id:
            warnings.warn('Current and target workspaces are the same')
            return

        self.current_ws_id = self.target_ws_id




import os

from .action_tracker import ActionTracker
from .bind import Bind
from .constants import _STATE
from .path_structure_meta import PathStructureMeta


class BasePathStructure(metaclass=PathStructureMeta):
    """Base class for the path structure"""
    
    def __init__(self, root: str, *, data=None):
        """Creates a path structure instance

        Parameters
        ----------
        root: str
            Path structure root path

        Examples
        --------

        :ref:`basic-path-structure`
        """
        root = str(root)
        self._root = root
        os.makedirs(self._root, exist_ok=True)

        self._state_path = os.path.join(root, _STATE)
        self._data = data

    def _bind_data(self, data):
        """Binds the data object"""
        assert data is not None
        self._data = data

    def track(self, action: str) -> ActionTracker:
        """Creates an ActionTracker object and targets action `action`
        
        It's a shortcut which creates an ActionTracker for you
        
        ..code-block:: python
        
            >>> from woger import BasePathStructure
            >>>
            >>> class PathStructure(BasePathStructure):
            ...     json = 'json'            
            >>>
            >>> ps = PathStructure('root')
            >>>
            >>> # WRONG
            >>> tracker = ActionTracker('load_json', 'root/state.json', ps.json)
            >>> 
            >>> # RIGHT
            >>> tracker = ps.track(ps.json.action())
            
        Parameters
        ----------
        action: str
            Action to track
            
        Returns
        -------
        ActionTracker instance
        """
        filename = str(Bind.from_action(action))
        path_to_check = os.path.join(self._root, filename)
        return ActionTracker(action, self._state_path, path=path_to_check)

    @classmethod
    def _path_attrs(cls):
        return list(cls._path_meta)

    def __repr__(self):
        return '<PathStructure {}>'.format(self._path_attrs())
